// Fetch-on-demand cache for texts.
//
// $Id: text-cache.cc,v 1.16 1995/11/18 22:50:14 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <stdlib.h>
#include "error-codes.h"
#include "text-cache.h"
#include "question.h"
#include "misc.h"
#include "async_text_created.h"
#include "uplink.h"

static Text_stat bad_text_stat;
static Text_stat zero_text_stat(KOM_TEXT_ZERO);

Text_stat_cache::Text_stat_cache(Connection *c,
				 Async_dispatcher *disp)
    : conn(c),
      async_disp(disp),
      cr_tag(disp->register_text_created_handler
	     (async_callback_method(this, &Text_stat_cache::new_text)))
{
}

Text_stat_cache::~Text_stat_cache()
{
    async_disp->unregister_text_created_handler(cr_tag);
}

bool
Text_stat_cache::present(Text_no cno) const
{
    return statusmap.present(cno);
}

#if 0
void
Text_stat_cache::clear()
{
    cerr << "Unimp\n";
    abort();
}
#endif

Connection *
Text_stat_cache::connection() const
{
    return conn;
}


Status
Text_stat_cache::prefetch(Text_no tno)
{
    if (tno <= 0)
	return st_error;

    if (statusmap.present(tno))
    {
	if (statusmap[tno].errno() == KOM_NO_ERROR)
	    return st_ok;
	else
	    return st_error;
    }
    else
    {
	if (!pendingmap.present(tno))
	{
	    pendingmap[tno] = new get_text_stat_question(conn, tno);
	    return st_pending;
	}
	else
	{
	    switch (pendingmap[tno]->status())
	    {
	    case st_pending:
		return st_pending;
	    case st_ok:
		statusmap[tno] = pendingmap[tno]->result();
		pendingmap.delete_entry(tno);
		return st_ok;
	    case st_error:
		statusmap[tno] = Text_stat(pendingmap[tno]->error());
		pendingmap.delete_entry(tno);
		return st_error;
	    }
	}
    }
    abort();
}


Text_stat
Text_stat_cache::get(Text_no tno)
{
    if (tno <= 0)
	return zero_text_stat;

    if (statusmap.present(tno))
    {
	// return statusmap[tno];
	Text_stat GCCBUG(statusmap[tno]);
	return GCCBUG;
    }
    else if (pendingmap.present(tno))
    {
	Text_stat res;
	get_text_stat_question *gtr = pendingmap[tno];
	
	if (gtr->receive() == st_ok)
	    res = gtr->result();
	else
	    res = Text_stat(gtr->error());
	statusmap[tno] = res;
	pendingmap.delete_entry(tno);
	return res;
    }
    else		
    {
	// Not present in cache, and no outstanding question.
	Text_stat res;
	get_text_stat_question gt(conn, tno);
	if (gt.receive() == st_ok)
	    res = gt.result();
	else
	{
	    res = Text_stat(gt.error());
	}
	statusmap[tno] = res;
	return res;
    }
    abort();
}

//void
//Text_stat_cache::remove(Text_no tno)
//{
//    abort();
//}

void
Text_stat_cache::new_text(const Async_text_created &msg)
{
    Text_stat t = msg.text_stat();
    statusmap[t.text_no()] = t;

    // New texts cannot have comments/footnotes.
    assert(t.num_comment_in() == 0);
    assert(t.num_footnote_in() == 0);

    // Update all texts that this one refers to.
    for (int ix = 0; ix < t.num_uplinks(); ix++)
    {
	Text_no parent = t.uplink(ix)->parent();
	if (statusmap.present(parent))
	{
	    Text_stat ts = statusmap[parent];
	    if (t.uplink(ix)->is_footnote())
		ts.add_footnote_in(t.text_no());
	    else
		ts.add_comment_in(t.text_no());
	}
    }
}

bool
Text_stat_cache::Tstmap::present(Text_no c) const
{
    // contains() really should be a const member function. This is a 
    // workaround. The line below should be substituted.
    return ((Text_stat_cache::Tstmap*)this)->contains(c) ? true : false;
    // return contains(c) ? true : false;
}

Text_stat &
Text_stat_cache::Tstmap::operator[](Text_no c)
{
    return Text_noText_statCHMap::operator[](c);
}


Text_stat_cache::Tstmap::Tstmap()
    : Text_noText_statCHMap(bad_text_stat, 50)
{
}

bool
Text_stat_cache::Tstreqmap::present(Text_no c)
{
    return contains(c) ? true : false;
}

void
Text_stat_cache::Tstreqmap::delete_entry(Text_no c)
{
    delete operator[](c);
    del(c);
}

get_text_stat_question *&
Text_stat_cache::Tstreqmap::operator[](Text_no c)
{
    return Text_noget_text_stat_question_ptrCHMap::operator[](c);
}

static get_text_stat_question_ptr bad_ptr = NULL;

Text_stat_cache::Tstreqmap::Tstreqmap()
    : Text_noget_text_stat_question_ptrCHMap(bad_ptr)
{
}

Text_stat_cache::Tstreqmap::~Tstreqmap()
{
    for (Pix i = first(); i != 0; next(i))
    {
	delete contents(i);
    }
}
