// Fetch-on-demand map from local to global text numbers.
//
// $Id: text-mapping.cc,v 1.17 1996/01/22 04:16:55 ceder Exp $
// Copyright (C) 1994  Per Cederqvist and Inge Wallin
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include <assert.h>
#include <stdlib.h>
#include <DLList.h>
#include "map-cache.h"
#include "kom-types.h"
#include "misc.h"
#include "conference.h"
#include "question.h"
#include "Local2Global.h"

Text_mapping::Interval::~Interval()
{
}

Text_mapping::Text_mapping()
    : mci(NULL)
{
}

Text_mapping::Text_mapping(Conference c, int chunk_size)
  : mci(new MCI(c))
{
    mci->add_ref();
    mci->chunk_size = chunk_size;
}

Text_mapping::Text_mapping(const Text_mapping&m)
  : mci(m.mci)
{
    if (mci)
	mci->add_ref();
}

Text_mapping& 
Text_mapping::operator=(const Text_mapping&m)
{
    if (this != &m)
    {
	if (mci)
	{
	    mci->sub_ref();
	    if (mci->refcount == 0)
		delete mci;
	}
	mci = m.mci;
	if (mci)
	    mci->add_ref();
    }
    return *this;
}

Text_mapping::~Text_mapping()
{
    if (mci)
    {
	mci->sub_ref();
	if (mci->refcount == 0)
	    delete mci;
    }
}

Pix
Text_mapping::locate_interval(Local_text_no lno)
{
    assert(mci != NULL);
    Pix ix;
    for (ix = mci->maps.first(); 
	 ix != 0 && mci->maps(ix)->first + mci->maps(ix)->chunk_size <= lno; 
	 mci->maps.next(ix)) 
    {
	continue;
    }

    if (ix == 0)
    {
	// Insert a new interval pointer last.
	mci->maps.append((Interval*)NULL);
	ix = mci->maps.last();
    }
    else if (mci->maps(ix)->first > lno)
    {
	mci->maps.ins_before(ix, (Interval*)NULL);
	mci->maps.prev(ix);
    }
    return ix;
}


void
Text_mapping::do_prefetch(Pix ix, 
			  Local_text_no lno,
			  int wanted_chunk_size)
{
    assert(mci != NULL);
    // We are linking in a new Unknown_interval into the map. First,
    // find out which interval we are going to ask about. We want it
    // to be wanted_chunk_size texts big, but we never want to ask
    // twice about a text. So, find out about the neighbouring
    // intervals.

    Pix iy = ix;
    Local_text_no min = mci->first_existing;
    mci->maps.prev(iy);
    if (iy)
    {
	assert(mci->maps(iy)->first + mci->maps(iy)->chunk_size > min);
	min = mci->maps(iy)->first + mci->maps(iy)->chunk_size;
    }

    iy = ix;
    Local_text_no max = mci->first_non_existing;
    mci->maps.next(iy);
    if (iy)
    {
	assert(mci->maps(iy)->first < max);
	max = mci->maps(iy)->first;
    }

    Local_text_no start = max - wanted_chunk_size;
    if (start < min)
	start = min;
    if (start > lno)
	start = lno;
    if (start + wanted_chunk_size > max)
	wanted_chunk_size = max - start;
    
    mci->maps(ix)->first = start;
    mci->maps(ix)->chunk_size = wanted_chunk_size;
    ((Unknown_interval*)(mci->maps(ix)))->mq 
	= new get_map_question(mci->conf.connection(), 
			       mci->conf.conf_no(),
			       start, wanted_chunk_size);
}

void
Text_mapping::fix_results()
{
    assert(mci != NULL);
    // Pass one: check for newly arrived results.
    Pix ix;
    for (ix = mci->maps.first(); ix != 0; mci->maps.next(ix)) 
    {
	if (mci->maps(ix)->valid() == false)
	{
	    get_map_question *q = ((Unknown_interval*)(mci->maps(ix)))->mq;
	    switch(q->status())
	    {
	    case st_pending:
		break;
	    case st_error:
		{
		    // +++ Need better error handling here.
		    Zero_interval *zp = new Zero_interval;
		    zp->first = mci->maps(ix)->first;
		    zp->chunk_size = mci->maps(ix)->chunk_size;
		    delete mci->maps(ix);
		    q = NULL;
		    mci->maps(ix) = zp;
		}
		break;
	    case st_ok:
		{
		    Known_interval *kp = new Known_interval;
		    Local2Global l2g = q->result();
		    kp->first = mci->maps(ix)->first;
		    kp->chunk_size = mci->maps(ix)->chunk_size;
		    delete mci->maps(ix);
		    q = NULL;
		    mci->maps(ix) = kp;
		    
		    assert(l2g.first() >= kp->first);
		    assert(l2g.size() <= kp->chunk_size);
		    
		    if (l2g.first() + l2g.size()
			< kp->first + kp->chunk_size)
		    {
			// We didn't get all we asked for, so we
			// must have gotten the last existing text.
			
			mci->first_non_existing = (l2g.first() 
						   + l2g.size());
		    }
		    
		    if (l2g.first() > kp->first)
		    {
			mci->first_existing = l2g.first();
			kp->first = l2g.first();
		    }
		    if (l2g.size() < kp->chunk_size)
		    {
			kp->chunk_size = l2g.size();
			// +++ Size might become zero here.
			// Does that matter?
		    }
		    kp->array = new Text_no[kp->chunk_size];

		    // +++ Should detect zeros in this loop.
		    for (int i=0; i < kp->chunk_size; i++)
			kp->array[i] = l2g[kp->first+i];
		}
		break;
	    default:
		abort();
	    }
	}
    }

    // Pass two: join joinable things.

    while ((!mci->maps.empty()) && mci->maps.front()->chunk_size == 0)
	mci->maps.del_front();

    ix = mci->maps.first();
    while (ix != 0)
    {
	Pix iy = ix;
	mci->maps.next(iy);
	if (iy == 0)
	    break;
	else if (mci->maps(iy)->chunk_size == 0)
	    mci->maps.del_after(ix);
	else if (mci->maps(ix)->first + mci->maps(ix)->chunk_size == 
		 mci->maps(iy)->first
		 && 0 /* s/0/is_joinable/ */ 
		 )
	{
	    // +++ Should join adjoining fields here.
	}
	else
	    mci->maps.next(ix);
    }
}

Text_no
Text_mapping::operator [](Local_text_no lno)
{
    if (mci==NULL)
	return 0;

    if (lno < mci->first_existing || mci->first_non_existing <= lno)
	return 0;

    Pix ix = locate_interval(lno);

    if (mci->maps(ix) == 0)
    {
	// Create a new interval.
	Unknown_interval *ui = new Unknown_interval;
	mci->maps(ix) = ui;
	do_prefetch(ix, lno, mci->chunk_size);
	return (*this)[lno];
    }
    else
    {
	if (mci->maps(ix)->valid())
	    return (*mci->maps(ix))[lno];
	((Unknown_interval*)(mci->maps(ix)))->mq->receive();
	fix_results();
	return (*this)[lno];
    }
}

Status
Text_mapping::prefetch(Local_text_no first,
		       int chunk_size)
{
    if (mci == NULL)
	return st_error;

    if (first < mci->first_existing || mci->first_non_existing <= first)
	return st_ok;

    Pix ix = locate_interval(first);

    if (mci->maps(ix) == 0)
    {
	// Create a new interval.
	Unknown_interval *ui = new Unknown_interval;
	mci->maps(ix) = ui;
	do_prefetch(ix, first, chunk_size);
	return st_pending;
    }
    else
    {
	if (mci->maps(ix)->valid())
	    return st_ok;
	else
	{
	    if (((Unknown_interval*)(mci->maps(ix)))->mq->status() 
		!= st_pending)
	    {
		fix_results();
		return st_ok;
	    }
	    else
		return st_pending;
	}
    }
}

void
Text_mapping::add(Local_text_no key, Text_no text)
{
    assert(mci != NULL);
    Pix ix = locate_interval(key);

    if (mci->maps(ix) == 0)
    {
	Known_interval *ki = new Known_interval;
	ki->first = key;
	ki->chunk_size = 1;
	ki->array = new Text_no[1];
	ki->array[0] = text;
	mci->maps(ix) = ki;
    }
    else
    {
	assert(mci->maps(ix)->first <= key
	       && mci->maps(ix)->first + mci->maps(ix)->chunk_size > key);
#if 0
	cerr << "Unexpected call to Text_mapping::add(Local_text_no, "
	    << "Text_no)\n";
	abort();
#endif
    }
}

void
Text_mapping::add(const Local2Global& l2g)
{
    assert(mci != NULL);
    for (Local_text_no l = l2g.first(); l < l2g.first() + l2g.size(); l++)
	if (l2g[l])
	    add(l, l2g[l]);
}


void
Text_mapping::limit()
{
    // +++ Not yet implemented.
}


Text_mapping::Unknown_interval::~Unknown_interval()
{
    if (mq)
        delete mq;
}

Text_mapping::Known_interval::~Known_interval()
{
    if (array)
        delete[] array;
}

Text_no
Text_mapping::Zero_interval::operator [](const Local_text_no lno) const
{
    assert(lno < first + chunk_size);
    return 0;
}

bool 
Text_mapping::Zero_interval::valid() const
{
    return true;
}

bool Text_mapping::Zero_interval::joinable() const
{
    return false;
}

Text_no
Text_mapping::Unknown_interval::operator [](const Local_text_no lno) const
{
    assert(lno < first + chunk_size);
    abort();
}

bool
Text_mapping::Unknown_interval::valid() const
{
    return false;
}

bool
Text_mapping::Unknown_interval::joinable() const
{
    return false;
}
	
Text_no
Text_mapping::Known_interval::operator [](const Local_text_no lno) const
{
    assert(lno < first + chunk_size);
    return array[lno - first];
}

bool 
Text_mapping::Known_interval::valid() const
{
    return true;
}

bool
Text_mapping::Known_interval::joinable() const
{
    return true;
}

#if 0
void
Text_mapping::Known_interval::join(const Known_interval &m)
{
    abort();
}
#endif

Text_mapping::MCI::MCI(Conference c)
  : conf(c), first_existing(0), first_non_existing((unsigned)-1),
    refcount(0),
    chunk_size(1)
{
}

//Text_mapping::MCI::MCI(const MCI&c)
//  : conf(c.conf)
//{
//    abort();
//}

//Text_mapping::MCI &
//Text_mapping::MCI::operator=(const MCI&)
//{
//    abort();
//}

Text_mapping::MCI::~MCI()
{
    for (Pix ix = maps.first(); ix != 0; maps.next(ix))
    {
	Interval *i = maps(ix);
	delete i;
    }
}

void
Text_mapping::MCI::add_ref()
{
    ++refcount;
}

void
Text_mapping::MCI::sub_ref()
{
    --refcount;
}

bool
Text_mapping::valid() const
{
    return (bool) !!mci;
}
