# Makefile.in generator.
# Copyright (C) 1996  Per Cederqvist
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

import string
import regsub
import types

"""Makefile.in generator module.

This module aids in the creation of Makefile.in files.  All that is
needed to create a Makefile.in that follows the GNU Coding Standards
is to create an ``makefile'' object, supply it with some data, and
call the ``emit_makefile'' method.

This file also contains the class ``komxxmakefile'', which is a
specialization of ``makefile'' with defaults suitable to the kom++
project.

TERMINOLOGY
-----------

Throughout this file, the following terminology is used:

- target: a Makefile target such as ``clean'', ``all'', ``foo.o'' or
  ``subdir-wwwkom-clean''.
- subdir: an immediate subdirectory, such as ``wwwkom'' (but the word
  subdir is not used for subdirs in subdirs, such as ``wwwkom/html'')
- feature: a ``configure.in'' variable, such as ``WWW_KOM'', that has
  one of the values ``yes'' or ``no''.  The ``configure.in'' script
  must use ``AC_SUBST'' to define it, and set it to the proper value.
  The value is typically derived from a ``--enable-FEATURE'', but it
  may also be computed by the ``configure.in'' script.
- phony target: a target that is not an actual file.  ``clean'' is
  typically a phony target, but ``foo.o'' is not.
- real target: a target that isn't phony.
- category: a class of real targets.  The category specifies
    - how the target is built (unless it comes with the distribution)
    - where it should be installed (if it should be installed)
    - when it should be removed (never, make clean, make distclean, et c)
- phase: one of ``source'', ``depend'', ``all'' and ``install''.
"""

# Exception
error='error'

phases = ["source", "depend", "all", "install"]
clean_phases = ["mostlyclean", "clean", "distclean", "realclean", "pure"]

standard_variables = """
# These are needed for compilation in separate directories.
srcdir = @srcdir@
top_srcdir = @top_srcdir@
VPATH = @srcdir@

# Just in case...
SHELL = /bin/sh

# See the Autoconf documentation for the exact purpose of these.
bindir = @bindir@
datadir = @datadir@
exec_prefix = @exec_prefix@
includedir = @includedir@
infodir = @infodir@
libdir = @libdir@
libexecdir = @libexecdir@
localstatedir = @localstatedir@
mandir = @mandir@
oldincludedir = @oldincludedir@
prefix = @prefix@
sbindir = @sbindir@
sharedstatedir = @sharedstatedir@
sysconfdir = @sysconfdir@

# Flags used when linking.
LDFLAGS = @LDFLAGS@
LIBS = @LIBS@

# Defines used by both the C and C++ compiler.
DEFS = @DEFS@

# C compiler setup.
CC = @CC@
CFLAGS = @CFLAGS@
ALL_CFLAGS = $(DEFS) $(CFLAGS) -I. -I@srcdir@ $(INCLUDES)
DEPENDFLAG = @DEPENDFLAG@

.c.o:
	$(CC) -c $(CPPFLAGS) $(ALL_CFLAGS) $<

# C++ compiler setup.
CXX = @CXX@
CXXFLAGS = @CXXFLAGS@
ALL_CXXFLAGS = $(DEFS) $(CXXFLAGS) -I. -I@srcdir@ $(INCLUDES)
XXDEPENDFLAG = @XXDEPENDFLAG@

.cc.o:
	$(CXX) -c $(CPPXXFLAGS) $(ALL_CXXFLAGS) $<

# Utility programs
AWK = @AWK@
INSTALL = @INSTALL@
INSTALL_DATA = @INSTALL_DATA@
INSTALL_PROGRAM = @INSTALL_PROGRAM@
MKDIR = mkdir
PURIFY = @PURIFY@
RANLIB = @RANLIB@
RM = rm -f
RUNTEST = runtest
RUNTESTFLAGS =
TOUCHLIB = @RANLIB@
"""


# Internal functions

def map_append(map, key, value):
    if not map.has_key(key):
	map[key] = []
    map[key].append(value)

class language:
    def __init__(self,language, depend_cmd):
	self.name = language
	self.depend_cmd = depend_cmd

language_cxx = language("C++", "$(CXX) $(XXDEPENDFLAG) $(ALL_CXXFLAGS)")
language_c = language("C", "$(CC) $(DEPENDFLAG) $(ALL_CFLAGS)")
		
class depend_registry:
    def __init__(self):
	self.languages = {}

    def register_language(self, language):
	self.languages[language.name] = language

    def create_depend_generators(self, items):
	srcs={}
	for i in items:
	    if self.languages.has_key(i.source_language()):
		map_append(srcs, i.source_language(), i.source())
	res=[]
	for lang in srcs.keys():
	    res.append(depend_generator(self.languages[lang].depend_cmd,
					srcs[lang]))
	return res

depender = depend_registry()
depender.register_language(language_cxx)
depender.register_language(language_c)

# Classes

class item:
    def __init__(self, name, features = None):
	self.name = name
	self.features = features

    def enabling_features(self):
	"""Return a list of enabling features, or None if unconditional."""
	return self.features

    def name_as_file(self):
	return self.name

    def name_as_target(self):
	# Remove any $, ( and ) characters.
	return regsub.gsub('[$()]*', '', self.name)

    def source_language(self):
	return None

class compiled_item(item):
    def source(self):
	return self.source_dir() \
	       + regsub.sub("\.o$", self.source_extension(), self.name)

    def source_dir(self):
	return "$(srcdir)/"

class generated_mixin:
    def source_dir(self):
	return "./"

class generated_item(generated_mixin, item):
    pass

class c_item(compiled_item):
    def source_extension(self):
	return ".c"

    def source_language(self):
	return "C"

class cxx_item(compiled_item):
    def source_extension(self):
	return ".cc"

    def source_language(self):
	return "C++"

class generated_cxx_item(generated_mixin, cxx_item):
    pass

class tcl_item(item):
    def source_language(self):
	return "TCL"

class generated_tcl_item(generated_mixin, tcl_item):
    pass
    
class conditional_container:
    def __init__(self, variable_prefix, variable_suffix):
	self.all = []
	self.conditional = []
	self.unconditional = []
	self.conditional_map = {}
	self.prefix = variable_prefix
	self.suffix = variable_suffix

    def variable(self):
	return string.upper(self.prefix + "_" + self.suffix)

    def cond_variable(self, feature, value=None):
	if value == None:
	    value = "@" + feature + "@"
	return string.upper(self.prefix) + "_" \
	       + feature + "_" + value + "_" \
	       + string.upper(self.suffix)

    def add_dep(self, dep):
	if type(dep) == types.StringType:
	    raise error, "string added"
	self.all.append(dep)
	if dep.enabling_features() == None:
	    self.unconditional.append(dep)
	else:
	    self.conditional.append(dep)
	    for feature in dep.enabling_features():
		map_append(self.conditional_map, feature, dep)

    def emit_variables(self):
	"""Emit all variable definitions this container needs."""
	if len(self.unconditional) > 0:
	    print self.variable(), "=",
	    for dep in self.unconditional:
		print dep.name_as_file(),
	    print
	for feature in self.conditional_map.keys():
	    print self.cond_variable(feature, "yes"), "=",
	    for dep in self.conditional_map[feature]:
		print dep.name_as_file(),
	    print
	    print self.cond_variable(feature, "no"), "="

    def emit_target_variables(self, target_prefix):
	for feature in self.conditional_map.keys():
	    print self.cond_variable(feature, "yes"), "=",
	    targets=[]
	    for dep in self.conditional_map[feature]:
		t = dep.name_as_target()
		if t not in targets:
		    targets.append(t)
	    for t in targets:
		print target_prefix + "-" + t,
	    print
	    print self.cond_variable(feature, "no"), "="
    
    def emit_deps(self):
	"""Emit all the dependencies.

	The dependencies are not output directly, but rather via the
	variables, so that conditional inclusion of dependencies
	works.
	"""
	if len(self.unconditional) > 0:
	    print "$(" + self.variable() + ")",
	self.emit_deps_conditional()

    def emit_deps_conditional(self):
	for feature in self.conditional_map.keys():
	    print "$(" + self.cond_variable(feature) + ")",

class install_data_mixin:
    def install_method(self):
	return "$(INSTALL_DATA)"

class target(install_data_mixin):
    """A target.

    Members:

    name:	the name of the target
    deps:	list objects that this target always requires.  (The
		objects are stored as strings).
    cond_deps:	a mapping from feature to list of objects that this
    		target require only if the feature is turned on.
    """

    # Target initialization

    def __init__(self):
	self.deps = conditional_container(self.category_name() + "_"
					  + self.name_upcase(),
					  self.category_collection_suffix())

    # Target information

    def is_real(self):
	"""Return nonzero if this is a real target.

	Real targets should, for instance, be deleted by ``make clean''.
	"""
	return 1

    def real_name(self):
	"""The name of the file this target builds.

	Usually equal to ``self.name'', but e.g. libraries override
	the method to perform transformations on the name.
	"""

	return self.name

    def category_name(self):
	raise error, "category_name not overridden for " \
	      + self.__class__.__name__

    def category_collection_suffix(self):
	"""The suffix used in the variable containing all deps."""
	raise error, "category_collection_suffix not overridden for " \
	      + self.__class__.__name__

    def name_upcase(self):
	return string.upper(self.name)

    # Target manipulation

    def add_dep(self, dep):
	self.deps.add_dep(dep)

    # Makefile.in output

    def emit_variables(self):
	"""Emit all variable definitions this target needs."""
	self.deps.emit_variables()

    def emit_rule_head(self):
	"""Emit the head of the rule that builds this target.

	Normally, this is the real name of the target, a colon, and
	all the dependencies.
	"""
	print self.real_name() + ":",
	self.emit_deps()
	print 

    def emit_deps(self):
	"""Emit all the dependencies.

	The dependencies are not output directly, but rather via the
	variables, so that conditional inclusion of dependencies
	works.
	"""
	self.deps.emit_deps()

    def emit_rule(self):
	"""Emit a rule that builds the target."""
	self.emit_rule_head()
	self.emit_rule_body()

    def emit_rule_body(self):
	"""Emit the body of the rule that builds the target.

	Each line output by this method must begin with a tab.
	"""
	raise error, "emit_rule_body not overridden for " \
	      + self.__class__.__name__

    # Installation

    def emit_install_rule(self):
	print self.install_target() + ": installdirs"
	print "\t" + self.install_method(), self.real_name(),
	print self.destination()

    def destination(self):
	raise error, "destination not overridden"

    def install_target(self):
	return "install-" + string.lower(self.category_name()) + "-"+ self.name

    # "make depend" support
    def depend_generator(self):
	# FIXME: Ignores conditional deps.
	return depender.create_depend_generators(self.deps.all)

    # "make clean" support
    def add_clean_info(self, makefile):
	if self.is_real():
	    makefile.mostlyclean(self.real_name())


class objs_mixin:
    def category_collection_suffix(self):
	return "OBJS"


class noinstall_mixin:
    def emit_install_rule(self):
	pass

    def install_target(self):
	return ""

    def destination(self):
	return None

class cxx_mixin:
    def add(self, object):
	self.add_dep(cxx_item(object))

    def add_generated(self, object):
	self.add_dep(generated_cxx_item(object))

class c_mixin:
    def add(self, object):
	self.add_dep(c_item(object))
    
class tcl_mixin:
    def add(self, object):
	self.add_dep(tcl_item(object))

class cxx_lib_target(noinstall_mixin, cxx_mixin, objs_mixin, target):
    def __init__(self, name):
	self.name = name
	target.__init__(self)

    def real_name(self):
	return "lib" + self.name + ".a"

    def emit_rule_body(self):
	print "\t$(AR) $(ARFLAGS) $@ $?"
	print "\t$(RANLIB) $@"

    def category_name(self):
	return "LIB"

class cc_lib_target(c_mixin, cxx_lib_target):
    pass

class bindir_mixin:
    def install_method(self):
	return "$(INSTALL_PROGRAM)"

    def destination(self):
	return "$(bindir)"

class sbindir_mixin(bindir_mixin):
    def destination(self):
	return "$(sbindir)"

class libexecdir_mixin(bindir_mixin):
    def destination(self):
	return "$(libexecdir)"


class shellscript_mixin:
    def category_name(self):
	return "SH"

    def category_collection_suffix(self):
	return "COLLECTION"
	
class sbin_shellscript(shellscript_mixin, sbindir_mixin, target):
    def __init__(self, name):
	self.name = name
	target.__init__(self)

    def emit_rule(self):
	pass


class bin_shellscript(shellscript_mixin, bindir_mixin, target):
    def __init__(self, name):
	self.name = name
	target.__init__(self)

    def emit_rule(self):
	pass


class libexec_shellscript(shellscript_mixin, libexecdir_mixin, target):
    def __init__(self, name):
	self.name = name
	target.__init__(self)

    def emit_rule(self):
	pass


class cxx_target(cxx_mixin, bindir_mixin, objs_mixin, target):
    def __init__(self, name):
	self.name = name
	target.__init__(self)
	self.libs = []

    def add_lib(self, lib):
	self.libs.append(lib)

    def emit_variables(self):
	"""Emit all variable definitions this target needs."""
	target.emit_variables(self)

    def emit_libs(self):
	for l in self.libs:
	    print l,

    def emit_rule_body(self):
	print "\t$(PURIFY) $(CXX) $(CXXLDFLAGS) $(ALL_CXXFLAGS) -o $@ \\"
	print "\t",
	self.emit_deps()
	print "\\\n\t",
	self.emit_libs()
	print

    def category_name(self):
	return "PROG"


class collection_target(target):
    def emit_rule(self):
	pass

    def is_real(self):
	return 0

    def category_name(self):
	return "COLLECTION"

    def category_collection_suffix(self):
	return "FILES"

    def emit_install_rule(self):
	print self.install_target() + ": installdirs"
	for f in self.deps.all:
	    print "\t" + self.install_method(), f.name_as_file(), \
		  self.destination()

class no_add_dep_mixin:
    def add_dep(self, dep):
	raise error, "add_dep called in class " + self.__class__.__name__

class generated_source(no_add_dep_mixin, noinstall_mixin, collection_target):
    def __init__(self, dep):
	self.name = dep
	collection_target.__init__(self)
	collection_target.add_dep(self, item(dep))

    def emit_variables(self):
	pass
	
    def emit_install_rule(self):
	pass

    def add_clean_info(self, makefile):
	makefile.mostlyclean(self.name)


class tcl_files(collection_target):
    def __init__(self):
	self.name = "tcl"
	collection_target.__init__(self)

    def destination(self):
	return "$(tcldir)"

class phase_target(noinstall_mixin, collection_target):
    """Used internally by the ``makefile'' class."""
    def __init__(self, phase):
	self.name = phase
	collection_target.__init__(self)

    def category_name(self):
	return "PHASE"

    def category_collection_suffix(self):
	return "TARGETS"

class depend_generator:
    def __init__(self, compiler, files):
	self.len = len(files)
	self.compiler = compiler
	self.files = string.join(files, " ")
	if self.len == 0:
	    raise error, "depend_generator with no files?"

    def emit_rule(self, output_file):
	if self.len == 0:
	    raise error, "Zero-length depend generator"
	print "\t", self.compiler, self.files, ">>" + output_file
    
class makefile:
    """Makefile.in generator class."""
    def __init__(self):
	self.user_targets = {}
	self.include_dirs = []
	self.my_targets = {}
	self.install_dirs = conditional_container("INSTALLDIR", "TARGET")
	self.install_targets = conditional_container("INSTALL", "TARGETS")
	self.init_clean_map()
	# Key: subdir name.  Value: List of (pair of target, list of
	# (pair of prerequisity directory and target)).
	self.subdirs = {}
	self.depend_generators = []

    def add_phase_target(self, phase, target_item):
	if not self.my_targets.has_key(phase):
	    self.my_targets[phase] = phase_target(phase)
	self.my_targets[phase].add_dep(target_item)

    def add_target(self, target, phase=None, feature=None):
	if target.name == None:
	    raise error, "Unnamed target"
	if self.user_targets.has_key(target.name):
	    raise error, "Respecification of target " + target.name
	if feature != None and len(feature) > 1:
	    raise error, "Too many features"
	self.user_targets[target.name] = target

	target.add_clean_info(self)

	if phase != None:
	    self.add_phase_target(phase, item(target.real_name(), feature))

	depend_gen = target.depend_generator()
	if type(depend_gen) != types.ListType:
	    raise error, "non-list depender"
	if depend_gen == None:
	    raise error, "no depender (impossible)"
	for d in depend_gen:
	    self.depend_generators.append(d)
	if len(depend_gen) > 0 and not self.my_targets.has_key("depend"):
	    self.my_targets["depend"] = phase_target("depend")
	# FIXME: handle feature != None in a better way.
	# No dependency generation should be performed
	# for the files that are only compiled conditionally.
	# Probably need to store all info about files in a
	# single dependency thingy with handling similar to
	# installdirs (handle each file separately).

	dest = target.destination()
	if dest != None:
	    if not self.my_targets.has_key("install"):
		self.my_targets["install"] = phase_target("install")

	    self.install_dirs.add_dep(item(dest, feature))
	    self.install_targets.add_dep(item(target.install_target(),
					      feature))

    def add_subdir(self, name, targetspeclist, featurelist):
	if self.subdirs.has_key(name):
	    raise error, "Subdir " + name + " added twice."
	self.subdirs[name] = (targetspeclist, featurelist)
	for spec in targetspeclist:
	    phase = spec[0]
	    if phase in phases:
		target = "descend-" + name + "-" + phase
		if featurelist == "always":
		    self.add_phase_target(phase, item(target))
		else:
		    self.add_phase_target(phase, item(target, featurelist))

    def add_include(self, dir):
	self.include_dirs.append(dir)

    def emit_banner(self):
	print "# Makefile.in generated by komxxmakefile.py"
	print "# Automatically generated -- do not edit!"
	print

    def emit_first_all(self):
	print "# Make sure that ``all'' is the default target."
	print "all:"
	print

    def emit_installdir_variables(self):
	self.install_dirs.emit_target_variables("installdir")

    def emit_install_features_variables(self):
	self.install_targets.emit_variables()

    def emit_std_variables(self):
	print standard_variables

    def emit_variables(self):
	self.emit_std_variables()
	self.emit_installdir_variables()
	self.emit_install_features_variables()
	for t in self.user_targets.values():
	    t.emit_variables()
	for t in self.my_targets.values():
	    t.emit_variables()
	if len(self.include_dirs) > 0:
	    print "INCLUDES =",
	    for i in self.include_dirs:
		print "-I" + i,
	    print

    def emit_build_rules(self):
	for t in self.user_targets.values():
	    t.emit_rule()
	print
	for t in self.my_targets.values():
	    t.emit_rule()

    def emit_install_rules(self):
	for t in self.user_targets.values():
	    t.emit_install_rule()
	print

    def init_clean_map(self):
	self.clean_map = {}
	self.clean_map["mostlyclean"] = ["*.o", "lib*.a", "core", "*.pyc"]
	self.clean_map["clean"] = ["stamp-*"]
	self.clean_map["distclean"] = ["*~", ".\#*", "TAGS", "Makefile"]
	self.clean_map["realclean"] = []
	self.clean_map["pure"] = ["Makefile.in"]

    def mostlyclean(self, file):
	self.clean_map["mostlyclean"].append(file)

    def clean(self, file):
	self.clean_map["clean"].append(file)
	
    def distclean(self, file):
	self.clean_map["distclean"].append(file)

    def realclean(self, file):
	self.clean_map["realclean"].append(file)

    def pure(self, file):
	self.clean_map["pure"].append(file)
	
    def emit_clean_rules(self):
	for target in clean_phases:
	    print target + ":",
	    for subtarget in clean_phases[:clean_phases.index(target)+1]:
		if len(self.clean_map[subtarget]) > 0:
		    print "do" + subtarget,
	    for subdir in self.subdirs.keys():
		print self.subcleantarget(subdir, target),
	    print
	for target in clean_phases:
	    if len(self.clean_map[target]) > 0:
		print "do" + target + ":"
		print "\t$(RM)",
		for file in self.clean_map[target]:
		    print file,
		print

    def emit_toplevel_install(self):
	print "doinstall: installdirs",
	self.install_targets.emit_deps()
	print

    def mkinstalldirs_program(self):
	return "mkinstalldirs"

    def emit_installdir_cmd(self, dirlist):
	print "\t" + self.mkinstalldirs_program(),
	for dir in dirlist:
	    print dir.name_as_file(),
	print

    def emit_installdirs(self):
	print "installdirs:",
	self.install_dirs.emit_deps_conditional()
	print
	if len(self.install_dirs.unconditional) > 0:
	    self.emit_installdir_cmd(self.install_dirs.unconditional)
	print
	dirs=[]
	names=[]
	for dir in self.install_dirs.conditional:
	    if dir.name_as_target() not in names:
		dirs.append(dir)
		names.append(dir.name_as_target())
	for dir in dirs:
	    print "installdir-" + dir.name_as_target() + ":"
	    self.emit_installdir_cmd([dir])
	    print

    def emit_rules(self):
	self.emit_build_rules()
	self.emit_install_rules()
	self.emit_clean_rules()
	self.emit_toplevel_rules()
	self.emit_toplevel_install()
	self.emit_installdirs()

    def emit_toplevel_rules(self):
	# FIXME: Detta r ett feltnk!  source, depend, all, och
	# install ska alltid finnas, ven om de inte alltid gr ngot.
	for phase in self.my_targets.keys():
	    target = self.my_targets[phase]

	    if not target.name in phases:
		raise error, "Strange target" + target.name

	    # Emit something like "doall: nilkom"
	    print "do" + target.name + ":",
	    target.emit_deps()
	    print

	    # Emit something like "all: dosource; $(MAKE) doall"
	    if phase == "source":
		print "source: dosource"
	    elif phase == "depend":
		if self.my_targets.has_key("source"):
		    print "depend: dosource"
		    print "\t$(MAKE) dodepend"
		else:
		    print "depend: dodepend"
	    elif phase == "all":
		if self.my_targets.has_key("source"):
		    print "all: dosource"
		    if self.need_depend:
			print "\tif [ -f", self.depend_stamp(), "];",
			print "then $(MAKE) doall;\\"
			print "\telse $(MAKE) dodepend doall; fi"
		    else:
			print "\t$(MAKE) doall"
		else:
		    print "all:"
		    if self.need_depend:
			print "\tif [ -f", self.depend_stamp(), "];",
			print "then $(MAKE) doall;\\"
			print "\telse $(MAKE) dodepend doall; fi"
		    else:
			print "\t$(MAKE) doall"
	    elif phase == "install":
		if self.my_targets.has_key("all"):
		    print "install:"
		    print "\t$(MAKE) all"
		    print "\t$(MAKE) doinstall"
		elif self.need_depend:
		    print "install:"
		    print "\t$(MAKE) depend"
		    print "\t$(MAKE) doinstall"
		elif self.my_targets.has_key("source"):
		    print "install:"
		    print "\t$(MAKE) source"
		    print "\t$(MAKE) doinstall"
		else:
		    print "install: doinstall"

    def subcleantarget(self, subdir, phase):
	return "descend-" + subdir + "-" + phase

    def emit_subdir_clean(self):
	for subdir in self.subdirs.keys():
	    for phase in clean_phases:
		print self.subcleantarget(subdir, phase) + ":"
		print "\tif [ -f " + subdir + "/Makefile ]; then \\"
		print "\t    if (cd", subdir + "; $(MAKE)", phase \
		      +  ");then true;\\"
		print "\t    else exit 1;fi;else true;fi"

    def emit_subdir_rules(self):
	for subdir in self.subdirs.keys():
	    for desc in self.subdirs[subdir][0]:
		phase = desc[0]
		prerequisites = desc[1]
		print "descend-" + subdir + "-" + phase + ":",
		for req in prerequisites:
		    req_dir = req[0]
		    req_phase = req[1]
		    print "descend-" + req_dir + "-" + req_phase,
		print
		if phase in phases:
		    phase = "do" + phase
		elif phase in clean_phases or phase == "installdirs":
		    pass
		else:
		    raise error, "Unknown how to descend in phase " + phase
		cmd="cd " + subdir + "; $(MAKE) " + phase
		if phase == "dodepend":
		    print "\t[ -f", subdir + "/" + self.depend_stamp(), "]",
		    print "|| {", cmd, "; }"
		else:
		    print "\t" + cmd
		print

    def emit_subdirs(self):
	self.emit_subdir_clean()
	self.emit_subdir_rules()

    def emit_trailer(self):
	print "# Automatically generated -- do not edit!"
	print "# pmg-generated part of Makefile.in ends here."

    def emit(self):
	self.compile_depend()
	self.emit_banner()
	self.emit_first_all()
	self.emit_variables()
	self.emit_rules()
	self.emit_subdirs()
	self.emit_depend_rules()
	self.emit_trailer()

    # Depend handling

    def depend_file_name(self):
	return "dependencies"

    def depend_file_tmp(self):
	return self.depend_file_name() + ".tmp"

    def emit_depend_rules(self):
	if len(self.depend_generators) > 0:
	    print self.include_makefile(self.depend_file_name())
	    print self.depend_file_name() + ":"
	    print "\ttouch", self.depend_file_name()
	self.emit_depend_generation()

    def depend_stamp(self):
	return "stamp-depend"

    def emit_depend_generation(self):
	# FIXME: handle conditional files.
	print "dodepend:"
	if len(self.depend_generators) > 0:
	    print "\t$(RM)", self.depend_stamp(), self.depend_file_name(),
	    print self.depend_file_tmp()
	    for gen in self.depend_generators:
		gen.emit_rule(self.depend_file_tmp())
	    print "\tmv", self.depend_file_tmp(), self.depend_file_name()
	print "\ttouch", self.depend_stamp()

    def compile_depend(self):
	self.need_depend = 0
	if len(self.depend_generators) > 0:
	    self.clean(self.depend_file_name())
	    self.need_depend = 1
	if self.my_targets.has_key("depend"):
	    self.need_depend = 1

    # "make" incompatibilities

    def include_makefile(self, file):
	return "include " + file + "\n"

class komxxmakefile(makefile):
    def init_clean_map(self):
	makefile.init_clean_map(self)

	self.clean("stamp-*")
	self.mostlyclean("Makefile.in2")
	self.mostlyclean("*.incl")

    def mkinstalldirs_program(self):
	return "$(top_srcdir)/mkinstalldirs"

    def emit_std_variables(self):
	makefile.emit_std_variables(self)
	print "# The version of kom++ -- this is set in configure.in."
	print "VERSION = @KOMXX_VERSION@"
	print
	print "GENCLASS = genclass"
	print
	print "# Where to find Python include files"
	print "PY_INCLUDES = @PY_INCLUDES@"
	print
	print "# The location of the default LysKOM server"
	print "DEFAULT_SERVER = @default_server@"
	print "DEFAULT_PORT = @default_port@"
