/***********************************************************
Copyright 1991-1995 by Stichting Mathematisch Centrum, Amsterdam,
The Netherlands.
Copyright 1995 by Per Cederqvist, Signum Support, Linkoping, Sweden.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Stichting Mathematisch
Centrum or CWI not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior permission.

STICHTING MATHEMATISCH CENTRUM DISCLAIMS ALL WARRANTIES WITH REGARD TO
THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
FITNESS, IN NO EVENT SHALL STICHTING MATHEMATISCH CENTRUM BE LIABLE
FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

/* Komxx_session objects */
#include "session.h"
#include "uplink.h"
#include "unread-confs.h"
#include "create_text_q.h"

// This is C++, so we definitely have prototypes.
// Tell Python.h so by defining this symbol.
#define HAVE_PROTOTYPES

// This entire file needs to be within extern "C".
// This hackery is here to avoid messing up emacs auto-indentation features.
#define BEGINEXTERNC extern "C" {
#define ENDEXTERNC }
BEGINEXTERNC

#include "Py/Python.h"

// Objects-------------------------------------------------
static PyObject *error_class = NULL;
static PyObject *errno_base_class = NULL;
static PyObject *errno_classes[KOM_max_errno];
static PyObject *error_no_connection_class = NULL;
static PyObject *error_sequence_class = NULL;
static PyObject *error_no_time_class = NULL;
static PyObject *error_not_logged_in_class = NULL;

static PyObject *komxx_st_pending;
static PyObject *komxx_st_ok;
static PyObject *komxx_st_error;


#define SIZE(array) (sizeof(array)/sizeof((array)[0]))

// Error object support funcitons--------------------------

static PyObject *
create_class(PyObject *base,
	     PyObject *module_dict,
	     const char *name,
	     PyObject **class_dict = NULL)
{
    LyStr nm("komxx.");
    nm << LyStr(name);
    const char *buf = nm.view_block();
    PyObject *d = PyDict_New();
    PyObject *s = PyString_FromString((char*)buf);
    nm.view_done(buf);
    PyObject *tuple;
    if (base == NULL)
	tuple = NULL;
    else
    {
	tuple = PyTuple_New(1);
	Py_INCREF(base);
	PyTuple_SetItem(tuple, 0, base);
    }
    PyObject *new_class = PyClass_New(tuple, d, s);
    Py_XDECREF(tuple);
    if (class_dict != NULL)
    {
	assert(*class_dict == NULL);
	*class_dict = d;
    }
    else
    {
	Py_DECREF(d);
    }
    Py_DECREF(s);
    PyDict_SetItemString(module_dict, (char*)name, new_class);
    return new_class;
}

static void
create_errno(PyObject *base,
	     PyObject *module_dict,
	     Kom_err eno,
	     const char *name,
	     const char *explanation)
{
    PyObject *dict = NULL;
    errno_classes[eno] = create_class(base, module_dict, name, &dict);
    PyObject *number = PyInt_FromLong(eno);
    PyObject *str = PyString_FromString((char*)explanation);
    PyDict_SetItemString(dict, "errno", number);
    PyDict_SetItemString(dict, "errmsg", str);
    Py_DECREF(number);
    Py_DECREF(str);
    Py_DECREF(dict);
}

static void
throw_exception(PyObject *except)
{
    PyObject *exc = PyInstance_New(except, NULL, NULL);
    PyErr_SetObject(except, exc);
    Py_DECREF(exc);
}

static void
init_errors(PyObject *md)	// module dictionary
{
    int size = SIZE(errno_classes);
    for (int e = 0; e < size; e++)
	errno_classes[e] = NULL;

    error_class = create_class(NULL, md, "error");

    PyObject *base;
    
    base = error_class;
    error_sequence_class = create_class(base, md, "sequence_error");
    errno_base_class = create_class(base, md, "errno_error");
    error_no_connection_class = create_class(base, md, "no_connection");

    base = error_sequence_class;
    error_no_time_class = create_class(base, md, "no_time_error");
    error_not_logged_in_class = create_class(base, md, "not_logged_in_error");

    base = errno_base_class;
    create_errno(base, md, KOM_OBSOLETE, "err_obsolete",
		 "Server call no longer implemented");
    create_errno(base, md, KOM_PWD, "err_pwd",
		 "Password is wrong or contains illegal chars");
    create_errno(base, md, KOM_LONG_STR, "err_long_str",
		 "String too long");
    create_errno(base, md, KOM_LOGIN, "err_login",
		 "Not logged in.");
    create_errno(base, md, KOM_LOGIN_DISALLOWED, "err_login_disallowed",
		 "Logins are currently not allowed");
    create_errno(base, md, KOM_CONF_ZERO, "err_conf_zero",
		 "Internal kom++ error: attempt to use conference number 0.");
    create_errno(base, md, KOM_UNDEF_CONF, "err_undef_conf",
		 "Undefined or secret conference");
    create_errno(base, md, KOM_UNDEF_PERS, "err_undef_pers",
		 "Undefined or secret person");
    create_errno(base, md, KOM_ACCESS, "err_access",
		 "No 'read/write permission'");
    create_errno(base, md, KOM_PERM, "err_perm",
		 "No permission");
    create_errno(base, md, KOM_NOT_MEMBER, "err_not_member",
		 "Not member in conf");
    create_errno(base, md, KOM_NO_SUCH_TEXT, "err_no_such_text",
		 "Text does not exist or is secret");
    create_errno(base, md, KOM_TEXT_ZERO, "err_text_zero",
		 "Can't use text no 0");
    create_errno(base, md, KOM_NO_SUCH_LOCAL_TEXT, "err_no_such_local_text",
		 "No such local text number");
    create_errno(base, md, KOM_LOCAL_TEXT_ZERO, "err_local_text_zero",
		 "Can't use local text number 0");
    create_errno(base, md, KOM_BAD_NAME, "err_bad_name",
		 "Bad name: too long or contains illegal chars");
    create_errno(base, md, KOM_INDEX_OUT_OF_RANGE, "err_index_out_of_range",
		 "Index out of range");
    create_errno(base, md, KOM_CONF_EXISTS, "err_conf_exists",
		 "Conference already exists");
    create_errno(base, md, KOM_PERS_EXISTS, "err_pers_exists",
		 "Person already exists");
    create_errno(base, md, KOM_SECRET_PUBLIC, "err_secret_public",
		 "Cannot be secret and public");
    create_errno(base, md, KOM_LETTER_BOX, "err_letter_box",
		 "Cannot change letter_box flag");
    create_errno(base, md, KOM_ILL_MISC, "err_ill_misc",
		 "Illegal misc field.");
    create_errno(base, md, KOM_ILLEGAL_INFO_TYPE, "err_illegal_info_type",
		 "Internal error: info_type parameter was illegal.");
    create_errno(base, md, KOM_ALREADY_RECIPIENT, "err_already_recipient",
		 "Already recipient to this text.");
    create_errno(base, md, KOM_ALREADY_COMMENT, "err_already_comment",
		 "Already comment to this text.");
    create_errno(base, md, KOM_ALREADY_FOOTNOTE, "err_already_footnote",
		 "Already footnote to this text.");
    create_errno(base, md, KOM_NOT_RECIPIENT, "err_not_recipient",
		 "Not recipient");
    create_errno(base, md, KOM_NOT_COMMENT, "err_not_comment",
		 "Not comment to this text.");
    create_errno(base, md, KOM_NOT_FOOTNOTE, "err_not_footnote",
		 "Not footnote to this text.");
    create_errno(base, md, KOM_RECIPIENT_LIMIT, "err_recipient_limit",
		 "Too many recipients");
    create_errno(base, md, KOM_COMM_LIMIT, "err_comm_limit",
		 "Too many comments");
    create_errno(base, md, KOM_FOOT_LIMIT, "err_foot_limit",
		 "Too many footnotes");
    create_errno(base, md, KOM_MARK_LIMIT, "err_mark_limit",
		 "Too many marks.");
    create_errno(base, md, KOM_NOT_AUTHOR, "err_not_author",
		 "Only the author may add footnotes or delete texts.");
    create_errno(base, md, KOM_UNDEF_SESSION, "err_undef_session",
		 "No such session exists.");
    create_errno(base, md, KOM_REGEX_ERROR, "err_regex_error",
		 "Regexp compilation failed.");
}

static void
set_errno_error(int err)
{
    PyObject *cls;		// exception class object
    PyObject *exc;		// exception instance object

    if (errno_classes[err] != NULL)
    {
	cls = errno_classes[err];
	exc = PyInstance_New(cls, NULL, NULL);
    }
    else
    {
	cls = errno_base_class;
	exc = PyInstance_New(cls, NULL, NULL);
	PyObject *number = PyInt_FromLong(err);
	LyStr s("unknown error number ");
	s << err;
	const char *buf = s.view_block();
	PyObject *str = PyString_FromString((char*)buf);
	s.view_done(buf);
	// FIXME: Is there a cleaner way to do this?
	// I don't feel comfortable accessing the cl_dict
	// member.
	PyDict_SetItemString(((PyClassObject*)exc)->cl_dict,
			     "errno", number);
	PyDict_SetItemString(((PyClassObject*)exc)->cl_dict,
			     "errmsg", str);
	Py_DECREF(number);
	Py_DECREF(str);
    }
    PyErr_SetObject(cls, exc);
    Py_DECREF(exc);
}

// Python Types--------------------------------------------

typedef struct {
    PyObject_HEAD
    Connection *connection;
    Session *session;
    Unread_confs *unread_confs;
} komxx_session_object;

typedef struct {
    PyObject_HEAD
    Text *data;
} komxx_text_object;

typedef struct {
    PyObject_HEAD
    Conference *data;
} komxx_conf_object;

typedef struct {
    PyObject_HEAD
    Person *data;
} komxx_pers_object;

typedef struct {
    PyObject_HEAD
    Uplink *data;
} komxx_uplink_object;

typedef struct {
    PyObject_HEAD
    Recipient *data;
} komxx_recipient_object;

typedef struct {
    PyObject_HEAD
    Conf_type *data;
} komxx_conf_type_object;

typedef struct {
    PyObject_HEAD
    Conf_z_info *data;
} komxx_conf_z_info_object;

// Support functions---------------------------------------

ENDEXTERNC

static PyObject *
convert_to_python(Status s)
{
    switch(s)
    {
    case st_pending:
	Py_INCREF(komxx_st_pending);
	return komxx_st_pending;
    case st_ok:
	Py_INCREF(komxx_st_ok);
	return komxx_st_ok;
    case st_error:
	Py_INCREF(komxx_st_error);
	return komxx_st_error;
    }
    abort();
}

static PyObject *
convert_to_python(bool b)
{
    if (b)
    {
	Py_INCREF(Py_True);
	return Py_True;
    }
    else
    {
	Py_INCREF(Py_False);
	return Py_False;
    }
}

static PyObject *
convert_to_python(struct tm t)
{
    if (t.tm_wday == 0)
	t.tm_wday = 6;
    else
	t.tm_wday--;

    t.tm_year += 1900;
    t.tm_mon++;
    t.tm_yday++;

    return Py_BuildValue("(lllllllll)",
			 (long)t.tm_year, (long)t.tm_mon, (long)t.tm_mday,
			 (long)t.tm_hour, (long)t.tm_min, (long)t.tm_sec,
			 (long)t.tm_wday, (long)t.tm_yday, (long)t.tm_isdst);
}

static PyObject *
convert_to_python(unsigned long l)
{
    return Py_BuildValue("l", l);
}

static PyObject *
convert_to_python(long l)
{
    return Py_BuildValue("l", l);
}

static PyObject *
convert_to_python(int i)
{
    return Py_BuildValue("i", i);
}

static PyObject *
convert_to_python(unsigned short s)
{
    return Py_BuildValue("i", (int)s);
}

static PyObject *
convert_to_python(const LyStr &str)
{
    const char *buf = str.view_block();
    PyObject *rv = Py_BuildValue("s#", buf, str.strlen());
    str.view_done(buf);
    return rv;
}

static PyObject *
convert_to_python(const Unread_info &info)
{
    PyObject *tuple = PyTuple_New(3);
    if (tuple == NULL) 
	return NULL;

    PyObject *conf = convert_to_python(info.conf);
    PyObject *minu = convert_to_python(info.min_unread);
    PyObject *maxu = convert_to_python(info.max_unread);

    if (conf == NULL || minu == NULL || maxu == NULL)
    {
	Py_XDECREF(conf);
	Py_XDECREF(minu);
	Py_XDECREF(maxu);
	Py_DECREF(tuple);
	return NULL;
    }
    int r = 0;
    r |= PyTuple_SetItem(tuple, 0, conf);
    r |= PyTuple_SetItem(tuple, 1, minu);
    r |= PyTuple_SetItem(tuple, 2, maxu);
    if (r != 0)
    {
	Py_DECREF(tuple);
	return NULL;
    }
    return tuple;
}

static PyObject *
convert_to_python(const Who_info &info)
{
    PyObject *tuple = PyTuple_New(5);
    if (tuple == NULL) 
	return NULL;

    PyObject *pers = convert_to_python(info.pers_no());
    PyObject *work = convert_to_python(info.working_conf());
    PyObject *sess = convert_to_python(info.session());
    PyObject *doin = convert_to_python(info.doing());
    PyObject *unam = convert_to_python(info.username());

    if (pers == NULL || work == NULL ||
	sess == NULL || doin == NULL || unam == NULL)
    {
	Py_XDECREF(pers);
	Py_XDECREF(work);
	Py_XDECREF(sess);
	Py_XDECREF(doin);
	Py_XDECREF(unam);
	Py_DECREF(tuple);
	return NULL;
    }
    int r = 0;
    r |= PyTuple_SetItem(tuple, 0, pers);
    r |= PyTuple_SetItem(tuple, 1, work);
    r |= PyTuple_SetItem(tuple, 2, sess);
    r |= PyTuple_SetItem(tuple, 3, doin);
    r |= PyTuple_SetItem(tuple, 4, unam);
    if (r != 0)
    {
	Py_DECREF(tuple);
	return NULL;
    }
    return tuple;
}

static PyObject *
convert_to_python(const Mark &mark)
{
    PyObject *tuple = PyTuple_New(2);
    if (tuple == NULL) 
	return NULL;

    PyObject *text = convert_to_python(mark.text_no());
    PyObject *type = convert_to_python(mark.type());

    if (text == NULL || type == NULL)
    {
	Py_XDECREF(text);
	Py_XDECREF(type);
	Py_DECREF(tuple);
	return NULL;
    }
    int r = 0;
    r |= PyTuple_SetItem(tuple, 0, text);
    r |= PyTuple_SetItem(tuple, 1, type);
    if (r != 0)
    {
	Py_DECREF(tuple);
	return NULL;
    }
    return tuple;
}

static PyObject *
convert_to_python(const Read_item &item)
{
    PyObject *tuple = PyTuple_New(2);
    if (tuple == NULL) 
	return NULL;

    PyObject *text = convert_to_python(item.text_no);
    PyObject *level = convert_to_python(item.nesting_level);

    if (text == NULL || level == NULL)
    {
	Py_XDECREF(text);
	Py_XDECREF(level);
	Py_DECREF(tuple);
	return NULL;
    }
    int r = 0;
    r |= PyTuple_SetItem(tuple, 0, text);
    r |= PyTuple_SetItem(tuple, 1, level);
    if (r != 0)
    {
	Py_DECREF(tuple);
	return NULL;
    }
    return tuple;
}

template <class T>
static PyObject *
convert_to_python(const Array<T> &arr)
{
    PyObject *res = PyList_New(arr.size());
    if (res == NULL)
	return NULL;

    for (int i = 0; i < arr.size(); i++)
    {
	PyObject *tuple = convert_to_python(arr[i]);
	
	if (PyList_SetItem(res, i, tuple))
	{
	    Py_DECREF(res);
	    return NULL;
	}
    }

    return res;
}

static PyObject *
convert_to_python(const Conf_type &t);

BEGINEXTERNC

#include "selector.incl"
#include "selector-assertion.incl"
#include "selector-errno.incl"


// Komxx_uplink <-> Python glue----------------------------------

static void
komxx_uplink_dealloc(komxx_uplink_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_uplink_methods[] = {
    {"is_footnote", komxx_uplink_is_footnote, 1, "No doc"},
    {"parent", komxx_uplink_parent, 1, "No doc"},
    {"sender", komxx_uplink_sender, 1, "No doc"},
    {"sent_later", komxx_uplink_sent_later, 1, "No doc"},
    {"sent_at", komxx_uplink_sent_at, 1, "No doc"},
    {NULL,		NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_uplink_getattr(komxx_uplink_object *self,
		     char *name)
{
    return Py_FindMethod(komxx_uplink_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_uplink_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"uplink",			/*tp_name*/
	sizeof(komxx_uplink_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_uplink_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_uplink_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM uplink object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_recipient <-> Python glue----------------------------------

static void
komxx_recipient_dealloc(komxx_recipient_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_recipient_methods[] = {
    {"recipient", komxx_recipient_recipient, 1, "No doc"},
    {"is_carbon_copy", komxx_recipient_is_carbon_copy, 1, "No doc"},
    {"local_no", komxx_recipient_local_no, 1, "No doc"},
    {"has_rec_time", komxx_recipient_has_rec_time, 1, "No doc"},
    {"rec_time", komxx_recipient_rec_time, 1, "No doc"},
    {"sender", komxx_recipient_sender, 1, "No doc"},
    {"has_sent_at", komxx_recipient_has_sent_at, 1, "No doc"},
    {"sent_at", komxx_recipient_sent_at, 1, "No doc"},
    {NULL,		NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_recipient_getattr(komxx_recipient_object *self,
			char *name)
{
    return Py_FindMethod(komxx_recipient_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_recipient_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"recipient",			/*tp_name*/
	sizeof(komxx_recipient_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_recipient_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_recipient_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM recipient object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_text methods-----------------------------------
// Exported methods.

static PyObject *
komxx_text_recipients(PyObject *s,
		      PyObject *args)
{
    komxx_text_object *self = (komxx_text_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (self->data->errno() != 0)
    {
	set_errno_error(self->data->errno());
	return NULL;
    }

    PyObject *res = PyList_New(self->data->num_recipients());
    for (int i = 0; i < self->data->num_recipients(); i++)
    {
	komxx_recipient_object *r;
	r = PyObject_NEW(komxx_recipient_object, &Komxx_recipient_type);
	r->data = new Recipient(*self->data->recipient(i));
	PyList_SetItem(res, i, (PyObject *)r);
    }
    
    return res;
}

static PyObject *
komxx_text_comments_in(PyObject *s,
		       PyObject *args)
{
    komxx_text_object *self = (komxx_text_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (self->data->errno() != 0)
    {
	set_errno_error(self->data->errno());
	return NULL;
    }

    PyObject *res = PyList_New(self->data->num_comment_in());
    for (int i = 0; i < self->data->num_comment_in(); i++)
    {
	PyObject *c = Py_BuildValue("i", self->data->comment_in(i));
	PyList_SetItem(res, i, c);
    }
    
    return res;
}

static PyObject *
komxx_text_footnotes_in(PyObject *s,
		       PyObject *args)
{
    komxx_text_object *self = (komxx_text_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (self->data->errno() != 0)
    {
	set_errno_error(self->data->errno());
	return NULL;
    }

    PyObject *res = PyList_New(self->data->num_footnote_in());
    for (int i = 0; i < self->data->num_footnote_in(); i++)
    {
	PyObject *c = Py_BuildValue("i", self->data->footnote_in(i));
	PyList_SetItem(res, i, c);
    }
    
    return res;
}

static PyObject *
komxx_text_uplinks(PyObject *s,
		      PyObject *args)
{
    komxx_text_object *self = (komxx_text_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (self->data->errno() != 0)
    {
	set_errno_error(self->data->errno());
	return NULL;
    }

    PyObject *res = PyList_New(self->data->num_uplinks());
    for (int i = 0; i < self->data->num_uplinks(); i++)
    {
	komxx_uplink_object *r;
	r = PyObject_NEW(komxx_uplink_object, &Komxx_uplink_type);
	r->data = new Uplink(*self->data->uplink(i));
	PyList_SetItem(res, i, (PyObject *)r);
    }
    
    return res;
}

// Komxx_text <-> Python glue----------------------------------

static void
komxx_text_dealloc(komxx_text_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_text_methods[] = {
	{"text_no",	  komxx_text_text_no,       1, "No doc"},
	{"prefetch",	  komxx_text_prefetch,      1, "No doc"},
	{"present",	  komxx_text_present,       1, "No doc"},
	{"exists",	  komxx_text_exists,        1, "No doc"},
	{"errno",	  komxx_text_errno,         1, "No doc"},
	{"creation_time", komxx_text_creation_time, 1, "No doc"},
	{"author",        komxx_text_author,        1, "No doc"},
	{"num_lines",	  komxx_text_num_lines,     1, "No doc"},
	{"num_chars",	  komxx_text_num_chars,     1, "No doc"},
	{"num_marks",	  komxx_text_num_marks,     1, "No doc"},
	{"recipients",    komxx_text_recipients,    1, "No doc"},
	{"comments_in",	  komxx_text_comments_in,   1, "No doc"},
	{"footnotes_in",  komxx_text_footnotes_in,  1, "No doc"},
	{"uplinks",	  komxx_text_uplinks,       1, "No doc"},
	{"subject",	  komxx_text_subject,       1, "No doc"},
	{"body",	  komxx_text_body,          1, "No doc"},
	{"mark_as_read",  komxx_text_mark_as_read,  1, "No doc"},
	{NULL,		NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_text_getattr(komxx_text_object *self,
		   char *name)
{
    return Py_FindMethod(komxx_text_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_text_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,				 /* ob_size*/
	"text",				 /* tp_name*/
	sizeof(komxx_text_object),	 /* tp_basicsize*/
	0,				 /* tp_itemsize*/
	/* methods */
	(destructor)komxx_text_dealloc,  /* tp_dealloc */
	0,				 /* tp_print */
	(getattrfunc)komxx_text_getattr, /* tp_getattr */
	0, 				 /* tp_setattr */
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0,			/*  tp_call */
	0, 			/*  tp_str */
	0, 			/*  tp_xxx1 */
	0, 			/*  tp_xxx2 */
	0, 			/*  tp_xxx3 */
	0, 			/*  tp_xxx4 */
	"LysKOM text object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_conf_type <-> Python glue----------------------------------

static void
komxx_conf_type_dealloc(komxx_conf_type_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_conf_type_methods[] = {
    {"is_protected", komxx_conf_type_is_protected, 1, "No doc"},
    {"is_original", komxx_conf_type_is_original, 1, "No doc"},
    {"is_secret", komxx_conf_type_is_secret, 1, "No doc"},
    {"is_letterbox", komxx_conf_type_is_letterbox, 1, "No doc"},
    {NULL,		NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_conf_type_getattr(komxx_conf_type_object *self,
		   char *name)
{
    return Py_FindMethod(komxx_conf_type_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_conf_type_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"conf_type",			/*tp_name*/
	sizeof(komxx_conf_type_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_conf_type_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_conf_type_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM conf_type object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_conf_z_info <-> Python glue----------------------------------

static void
komxx_conf_z_info_dealloc(komxx_conf_z_info_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_conf_z_info_methods[] = {
    {"conf_no", komxx_conf_z_info_conf_no, 1, "No doc"},
    {"conf_type", komxx_conf_z_info_conf_type, 1, "No doc"},
    {"conf_name", komxx_conf_z_info_conf_name, 1, "No doc"},
    {NULL,		NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_conf_z_info_getattr(komxx_conf_z_info_object *self,
		   char *name)
{
    return Py_FindMethod(komxx_conf_z_info_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_conf_z_info_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"conf_z_info",			/*tp_name*/
	sizeof(komxx_conf_z_info_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_conf_z_info_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_conf_z_info_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM conf_z_info object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_conf <-> Python glue----------------------------------

static void
komxx_conf_dealloc(komxx_conf_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_conf_methods[] = {
    {"prefetch", komxx_conf_prefetch, 1, "No doc"},
    // FIXME: use new uconf_get call, if available.
    {"prefetch_name", komxx_conf_prefetch, 1, "No doc"},
    {"errno", komxx_conf_errno, 1, "No doc"},
    {"present", komxx_conf_present, 1, "No doc"},
    {"exists", komxx_conf_exists, 1, "No doc"},
    {"conf_no", komxx_conf_conf_no, 1, "No doc"},
    {"creator", komxx_conf_creator, 1, "No doc"},
    {"creation_time", komxx_conf_creation_time, 1, "No doc"},
    {"presentation", komxx_conf_presentation, 1, "No doc"},
    {"supervisor", komxx_conf_supervisor, 1, "No doc"},
    {"permitted_submitters", komxx_conf_permitted_submitters, 1, "No doc"},
    {"super_conf", komxx_conf_super_conf, 1, "No doc"},
    {"type", komxx_conf_type, 1, "No doc"},
    {"last_written", komxx_conf_last_written, 1, "No doc"},
    {"msg_of_day", komxx_conf_msg_of_day, 1, "No doc"},
    {"nice", komxx_conf_nice, 1, "No doc"},
    {"name", komxx_conf_name, 1, "No doc"},
    {"no_of_members", komxx_conf_no_of_members, 1, "No doc"},
    {"first_local_no", komxx_conf_first_local_no, 1, "No doc"},
    {"no_of_texts", komxx_conf_no_of_texts, 1, "No doc"},
    {NULL, NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_conf_getattr(komxx_conf_object *self,
		   char *name)
{
    return Py_FindMethod(komxx_conf_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_conf_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"conf",			/*tp_name*/
	sizeof(komxx_conf_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_conf_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_conf_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM conf object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_pers <-> Python glue----------------------------------

static void
komxx_pers_dealloc(komxx_pers_object *self)
{
    delete self->data;
    PyMem_DEL(self);
}

static PyMethodDef komxx_pers_methods[] = {
    {"prefetch", komxx_pers_prefetch, 1, "No doc"},
    {"prefetch_conf_only", komxx_pers_prefetch_conf_only, 1, "No doc"},
    // FIXME: use new uconf_get call, if available.
    {"prefetch_name", komxx_pers_prefetch_conf_only, 1, "No doc"},
    {"prefetch_pers_only", komxx_pers_prefetch_pers_only, 1, "No doc"},
    {"exists", komxx_pers_exists, 1, "No doc"},
    {"present", komxx_pers_present, 1, "No doc"},
    {"conf_present", komxx_pers_conf_present, 1, "No doc"},
    {"pers_present", komxx_pers_pers_present, 1, "No doc"},
    {"errno", komxx_pers_errno, 1, "No doc"},
    {"pers_no", komxx_pers_pers_no, 1, "No doc"},
    {"user_area", komxx_pers_user_area, 1, "No doc"},
#if 0
    // FIXME
    {"privileges", komxx_pers_privileges, 1, "No doc"},
    {"flags", komxx_pers_flags, 1, "No doc"},
#endif	
    {"last_login", komxx_pers_last_login, 1, "No doc"},
    {"total_time_present", komxx_pers_total_time_present, 1, "No doc"},
    {"sessions", komxx_pers_sessions, 1, "No doc"},
    {"created_lines", komxx_pers_created_lines, 1, "No doc"},
    {"created_bytes", komxx_pers_created_bytes, 1, "No doc"},
    {"read_texts", komxx_pers_read_texts, 1, "No doc"},
    {"no_of_text_fetches", komxx_pers_no_of_text_fetches, 1, "No doc"},
    {"no_of_marks", komxx_pers_no_of_marks, 1, "No doc"},
    {"no_of_confs", komxx_pers_no_of_confs, 1, "No doc"},
    {"creator", komxx_pers_creator, 1, "No doc"},
    {"creation_time", komxx_pers_creation_time, 1, "No doc"},
    {"presentation", komxx_pers_presentation, 1, "No doc"},
    {"supervisor", komxx_pers_supervisor, 1, "No doc"},
    {"permitted_submitters", komxx_pers_permitted_submitters, 1, "No doc"},
    {"super_conf", komxx_pers_super_conf, 1, "No doc"},
    {"last_written", komxx_pers_last_written, 1, "No doc"},
    {"msg_of_day", komxx_pers_msg_of_day, 1, "No doc"},
    {"nice", komxx_pers_nice, 1, "No doc"},
    {"name", komxx_pers_name, 1, "No doc"},
    {"no_of_members", komxx_pers_no_of_members, 1, "No doc"},
    {"first_local_no", komxx_pers_first_local_no, 1, "No doc"},
    {"no_of_texts", komxx_pers_no_of_texts, 1, "No doc"},
    {NULL, NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_pers_getattr(komxx_pers_object *self,
		   char *name)
{
    return Py_FindMethod(komxx_pers_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_pers_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"pers",			/*tp_name*/
	sizeof(komxx_pers_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_pers_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_pers_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM pers object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Komxx_session methods-----------------------------------

// Helper function: make sure that unread_confs has been fetched.
static Status
require_unread_confs(komxx_session_object *self)
{
    if (self->unread_confs == NULL)
    {
	if (self->session->my_login() == 0)
	{
	    throw_exception(error_not_logged_in_class);
	    return st_error;
	}
	self->unread_confs = new Unread_confs(self->session,
					      self->session->my_login());
    }

    return st_ok;
}

// Exported methods.

static PyObject *
komxx_session_create_person(PyObject *s,
			    PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    Pers_no pno = 0;
    char *name;
    int name_len;
    char *pwd;
    int pwd_len;

    if (!PyArg_ParseTuple(args, "s#s#", &name, &name_len, &pwd, &pwd_len))
	return NULL;

    // FIXME: This loses the autologin flag.
    switch (self->session->create_person(LyStr(name, name_len),
					 LyStr(pwd, pwd_len), &pno))
    {
    case st_ok:
	return convert_to_python(pno);
    case st_error:
	return convert_to_python(st_error);
    default:
	abort();
    }
}

static PyObject *
komxx_session_login(PyObject *s,
		    PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int pno = 0;
    char *pwd = NULL;
    int pwd_len = 0;
    int invisible = 0;

    if (!PyArg_ParseTuple(args, "is#|i", &pno, &pwd, &pwd_len, &invisible))
	return NULL;

    switch (self->session->login(pno, LyStr(pwd, pwd_len), invisible))
    {
    case st_ok:
	return convert_to_python(st_ok);
    case st_error:
	return convert_to_python(st_error);
    default:
	abort();
    }
}

static PyObject *
komxx_session_logout(PyObject *s,
		    PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (self->unread_confs != NULL)
    {
	delete self->unread_confs;
	self->unread_confs = NULL;
    }
    
    return convert_to_python(self->session->logout());
}

static PyObject *
komxx_session_add_member(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int cno;
    int reader;
    int prio;
    int place;

    if (!PyArg_ParseTuple(args, "iiii", &cno, &reader, &prio, &place))
	return NULL;

    if (reader < 0 || cno < 0 || prio < 0 || place < 0
	|| prio > 255 || place > 255)
    {
	PyErr_SetString(PyExc_ValueError, "Integer out of range");
	return NULL;
    }

    if (reader == self->session->my_login())
    {
	// This code is almost verbatimly duplicated in
	// komxx_session_join_conference.
	if (require_unread_confs(self) != st_ok)
	    return NULL;

	return convert_to_python(
	    self->unread_confs->join_conference(cno, prio, place));
    }
    else 
    {
	return convert_to_python(self->session->add_member(cno, reader,
							   prio, place));
    }
}

static PyObject *
komxx_session_send_message(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int rcpt;
    char *msg;
    int msg_len;

    if (!PyArg_ParseTuple(args, "is#", &rcpt, &msg, &msg_len))
	return NULL;

    if (rcpt < 0)
    {
	PyErr_SetString(PyExc_ValueError, "Integer out of range");
    }

    return convert_to_python(self->session->send_message(rcpt,
							 LyStr(msg, msg_len)));
}

static PyObject *
komxx_session_change_what_i_am_doing(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    char *doing;
    int doing_len;

    if (!PyArg_ParseTuple(args, "s#", &doing, &doing_len))
	return NULL;

    return convert_to_python(
	self->session->change_what_i_am_doing(LyStr(doing, doing_len)));
}

static PyObject *
komxx_session_is_member(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int pers;
    int conf;

    if (!PyArg_ParseTuple(args, "ii", &pers, &conf))
	return NULL;

    return convert_to_python(
	self->session->is_member(pers, conf));
}

static PyObject *
komxx_session_create_text(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    char *subject;
    int subject_len;
    char *body;
    int body_len;
    PyObject *recips;
    PyObject *cc_recips;
    PyObject *comment_to;
    PyObject *footnote_to;

    if (!PyArg_ParseTuple(args, "s#s#O!O!O!O!",
			  &subject, &subject_len, &body, &body_len,
			  &PyList_Type, &recips, 
			  &PyList_Type, &cc_recips, 
			  &PyList_Type, &comment_to, 
			  &PyList_Type, &footnote_to))
	return NULL;

    DLList<Recipient> rec;
    DLList<Text_no> com;
    DLList<Text_no> foot;

    int i;
    for (i = 0; i < PyList_Size(recips); i++)
    {
	PyObject *r = PyList_GetItem(recips, i);
	long recip = PyInt_AsLong(r);
	if (PyErr_Occurred())
	    return NULL;
	rec.append(Recipient((Conf_no)recip, false));
    }
    for (i = 0; i < PyList_Size(cc_recips); i++)
    {
	PyObject *r = PyList_GetItem(cc_recips, i);
	long recip = PyInt_AsLong(r);
	if (PyErr_Occurred())
	    return NULL;
	rec.append(Recipient((Conf_no)recip, true));
    }
    for (i = 0; i < PyList_Size(comment_to); i++)
    {
	PyObject *r = PyList_GetItem(comment_to, i);
	long tno = PyInt_AsLong(r);
	if (PyErr_Occurred())
	    return NULL;
	com.append((Text_no)tno);
    }
    for (i = 0; i < PyList_Size(footnote_to); i++)
    {
	PyObject *r = PyList_GetItem(footnote_to, i);
	long tno = PyInt_AsLong(r);
	if (PyErr_Occurred())
	    return NULL;
	foot.append((Text_no)tno);
    }
    LyStr text_mass(subject, subject_len);
    text_mass << '\n' << LyStr(body, body_len);
    create_text_question q(self->connection, text_mass, rec, com, foot);
    if (q.receive() == st_error)
    {
	set_errno_error(q.error());
	return NULL;
    }
    else
	return convert_to_python(q.result());
}


static PyObject *
komxx_session_lookup_z_name(PyObject *s,
			    PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    char *templ = NULL;
    int templ_len = 0;
    int want_persons = 1;
    int want_confs = 1;

    if (!PyArg_ParseTuple(args, "s#|ii", &templ, &templ_len,
			  &want_persons, &want_confs))
	return NULL;

    // FIXME: error handling.    
    Array<Conf_z_info> matches(self->session
			       ->lookup_z_name(LyStr(templ, templ_len),
					       want_persons, want_confs));

    PyObject *res = PyList_New(matches.size());
    for (int i = 0; i < matches.size(); i++)
    {
	komxx_conf_z_info_object *r;
	r = PyObject_NEW(komxx_conf_z_info_object, &Komxx_conf_z_info_type);
	r->data = new Conf_z_info(matches[i]);
	PyList_SetItem(res, i, (PyObject *)r);
    }

    return res;
}

static PyObject *
komxx_session_text(PyObject *s,
		   PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;

    int tno = 0;
    if (!PyArg_ParseTuple(args, "i", &tno))
	return NULL;

    komxx_text_object *rv = PyObject_NEW(komxx_text_object, &Komxx_text_type);
    if (rv == NULL)
	return NULL;

    rv->data = new Text(self->session->texts()[tno]);
    return (PyObject*)rv;
}

static PyObject *
komxx_session_conf(PyObject *s,
		   PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;

    int cno = 0;
    if (!PyArg_ParseTuple(args, "i", &cno))
	return NULL;

    komxx_conf_object *rv = PyObject_NEW(komxx_conf_object, &Komxx_conf_type);
    if (rv == NULL)
	return NULL;

    rv->data = new Conference(self->session->conferences()[cno]);
    return (PyObject*)rv;
}

static PyObject *
komxx_session_pers(PyObject *s,
		   PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;

    int pno = 0;
    if (!PyArg_ParseTuple(args, "i", &pno))
	return NULL;

    komxx_pers_object *rv = PyObject_NEW(komxx_pers_object, &Komxx_pers_type);
    if (rv == NULL)
	return NULL;

    rv->data = new Person(self->session->persons()[pno]);
    return (PyObject*)rv;
}

static PyObject *
komxx_session_prefetch(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int texts;
    int text_in_confs;
    int confs;
    int maps;
    int maps_in_confs;
    int map_confs;
    int max_question;

    Work_quota w;

    if (PyArg_ParseTuple(args, ""))
	w = Work_quota();
    else if (PyArg_ParseTuple(args, "iii", &maps, &texts, &max_question))
	w = Work_quota(maps, texts, max_question);
    else if (PyArg_ParseTuple(args, "iiiiiii", &texts, &text_in_confs, &confs,
			      &maps, &maps_in_confs, &map_confs,
			      &max_question))
	w = Work_quota(texts, text_in_confs, confs, maps, maps_in_confs,
		       map_confs, max_question);
    else
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->prefetch(&w));
}

static PyObject *
komxx_session_goto_next_conf(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->goto_next_conf());
}

static PyObject *
komxx_session_goto_conf(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int conf_no;

    if (!PyArg_ParseTuple(args, "i", &conf_no))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->goto_conf(conf_no));
}

static PyObject *
komxx_session_next_prompt(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;

    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->next_prompt());
}

static PyObject *
komxx_session_next_text(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int remove_it = 0;
    
    if (!PyArg_ParseTuple(args, "|i", &remove_it))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->next_text(remove_it));
}

static PyObject *
komxx_session_remove_text(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int tno;

    if (!PyArg_ParseTuple(args, "i", &tno))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    self->unread_confs->remove_text(tno);

    Py_INCREF(Py_None);
    return Py_None;
}

static PyObject *
komxx_session_current_conf(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->current_conf());
}

static PyObject *
komxx_session_confs_with_unread(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(self->unread_confs->confs_with_unread());
}

static PyObject *
komxx_session_min_unread_texts_in_current(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(
	self->unread_confs->min_unread_texts_in_current());
}

static PyObject *
komxx_session_max_unread_texts_in_current(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    return convert_to_python(
	self->unread_confs->max_unread_texts_in_current());
}

static PyObject *
komxx_session_list_unread_texts_in_current(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int wanted_texts = 100;
    
    if (!PyArg_ParseTuple(args, "|i", &wanted_texts))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    Read_order_iterator begin = self->unread_confs->current_conf_begin(wanted_texts);
    Read_order_iterator end = self->unread_confs->current_conf_end();
    Read_order_iterator iter;
    int size = 0;
    for (iter = begin; iter != end; ++iter)
	++size;

    Array<Read_item> arr(size);
    int n;
    for (iter = begin, n = 0; iter != end; ++iter, ++n)
	arr[n] = *iter;

    return convert_to_python(arr);
}

static PyObject *
komxx_session_join_conference(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int cno;
    int prio;
    int place;

    if (!PyArg_ParseTuple(args, "iii", &cno, &prio, &place))
	return NULL;

    if (require_unread_confs(self) != st_ok)
	return NULL;

    // The following code is almost verbatimly duplicated in
    // komxx_session_join_conference.
    if (cno < 0 || prio < 0 || place < 0 || prio > 255 || place > 255)
    {
	PyErr_SetString(PyExc_ValueError, "Integer out of range");
    }

    return convert_to_python(
	self->unread_confs->join_conference(cno, prio, place));
}

static PyObject *
komxx_session_block(PyObject *, PyObject *args)
{
    int timeout;

    if (!PyArg_ParseTuple(args, "i", &timeout))
	return NULL;

    return convert_to_python(Connection::drain(timeout));
}

static PyObject *
komxx_session_handle_async(PyObject *s, PyObject *args)
{
    komxx_session_object *self = (komxx_session_object*)s;
    int asyncs = 1000;

    if (!PyArg_ParseTuple(args, "|i", &asyncs))
	return NULL;

    return convert_to_python(self->session->handle_async(asyncs));
}



// Komxx_session <-> Python glue--------------------------------

static void
komxx_session_dealloc(komxx_session_object *self)
{
    delete self->unread_confs;
    delete self->session;
    delete self->connection;
    PyMem_DEL(self);
}

static PyMethodDef komxx_session_methods[] = {
    // Get prefetch-on-demand shadow objects.
    {"text",	komxx_session_text, 1, "No doc"},
    {"conf",	komxx_session_conf, 1, "No doc"},
    {"pers",	komxx_session_pers, 1, "No doc"},

    // State retrieval
    {"my_login", komxx_session_my_login, 1, "No doc"},
    {"my_session", komxx_session_my_session, 1, "No doc"},

    // Unread_confs handling
    {"prefetch",	komxx_session_prefetch, 1, "No doc"},
    {"goto_next_conf",	komxx_session_goto_next_conf, 1, "No doc"},
    {"goto_conf",	komxx_session_goto_conf, 1, "No doc"},
    {"next_prompt",	komxx_session_next_prompt, 1, "No doc"},
    {"next_text",	komxx_session_next_text, 1, "No doc"},
    {"remove_text",	komxx_session_remove_text, 1, "No doc"},
    {"current_conf",	komxx_session_current_conf, 1, "No doc"},
    {"confs_with_unread",	komxx_session_confs_with_unread, 1, "No doc"},
    {"min_unread_texts_in_current",
     komxx_session_min_unread_texts_in_current, 1, "No doc"},
    {"max_unread_texts_in_current",
     komxx_session_max_unread_texts_in_current, 1, "No doc"},
    {"list_unread_texts_in_current",
     komxx_session_list_unread_texts_in_current, 1, "No doc"},
    {"join_conference",	komxx_session_join_conference, 1, "No doc"},
    
    // Perform utility functions.
    {"block", komxx_session_block, 1, "No doc"},
    {"handle_async", komxx_session_handle_async, 1, "No doc"},
    {"create_person", komxx_session_create_person, 1, "No doc"},
    {"add_member", komxx_session_add_member, 1, "No doc"},
    {"send_message", komxx_session_send_message, 1, "No doc"},
    {"get_time", komxx_session_get_time, 1, "No doc"},
    {"change_what_i_am_doing",
     komxx_session_change_what_i_am_doing, 1, "No doc"},
    {"who_is_on", komxx_session_who_is_on, 1, "No doc"},
    {"get_marks", komxx_session_get_marks, 1, "No doc"},
    {"is_member", komxx_session_is_member, 1, "No doc"},
    {"create_text", komxx_session_create_text, 1, "No doc"},
#if 0
    {"handle_async", komxx_session_handle_async, 1, "No doc"},
#endif

    {"login",	komxx_session_login, 1, "No doc"},
    {"logout",	komxx_session_logout, 1, "No doc"},
    {"lookup_z_name", komxx_session_lookup_z_name, 1, "No doc"},
    {NULL,		NULL, 1, "end"}		/* sentinel */
};

static PyObject *
komxx_session_getattr(komxx_session_object *self,
		      char *name)
{
    return Py_FindMethod(komxx_session_methods, (PyObject *) self, name);
}

static PyTypeObject Komxx_session_type = {
	PyObject_HEAD_INIT(&PyType_Type)
	0,			/*ob_size*/
	"session",			/*tp_name*/
	sizeof(komxx_session_object),	/*tp_basicsize*/
	0,			/*tp_itemsize*/
	/* methods */
	(destructor)komxx_session_dealloc, /*tp_dealloc*/
	0,			/*tp_print*/
	(getattrfunc)komxx_session_getattr, /*tp_getattr*/
	0, /*tp_setattr*/
	0,			/*tp_compare*/
	0,			/*tp_repr*/
	0,			/*tp_as_number*/
	0,			/*tp_as_sequence*/
	0,			/*tp_as_mapping*/
	0,			/*tp_hash*/
	0, /*  tp_call */
	0, /*  tp_str */
	0, /*  tp_xxx1 */
	0, /*  tp_xxx2 */
	0, /*  tp_xxx3 */
	0, /*  tp_xxx4 */
	"LysKOM session object.\n\nDoc not yet available.", /*  *tp_doc */
};


// Global functions----------------------------------------------------

static PyObject *
komxx_session_new(PyObject *,
		  PyObject *args)
{
    char *server = NULL;
    char *port = NULL;

    if (!PyArg_ParseTuple(args, "ss", &server, &port))
	return NULL;

    komxx_session_object *rv = PyObject_NEW(komxx_session_object,
					    &Komxx_session_type);
    if (rv == NULL)
	return NULL;

    rv->connection = new Connection(server, port);
    if (rv->connection->connection_failed())
    {
	throw_exception(error_no_connection_class);
	return NULL;
    }

    rv->session = new Session(rv->connection);
    rv->unread_confs = NULL;
    return (PyObject *)rv;
}

// Global function glue-------------------------------------------------
static PyMethodDef komxx_functions[] = {
	{"new_session",		komxx_session_new, 1, "No doc available"},
	{NULL,		NULL, 1, "End"}		/* sentinel */
};

void
initkomxx()
{
    PyObject *m = Py_InitModule("komxx", komxx_functions);
    PyObject *d = PyModule_GetDict(m);
    init_errors(d);
    komxx_st_pending = PyString_FromString("komxx.st_pending");
    PyDict_SetItemString(d, "st_pending", komxx_st_pending);
    komxx_st_ok = PyString_FromString("komxx.st_ok");
    PyDict_SetItemString(d, "st_ok", komxx_st_ok);
    komxx_st_error = PyString_FromString("komxx.st_error");
    PyDict_SetItemString(d, "st_error", komxx_st_error);
    if (PyErr_Occurred())
	Py_FatalError("can't initialize module komxx");
}

ENDEXTERNC

static PyObject *
convert_to_python(const Conf_type &t)
{
    komxx_conf_type_object *r;
    r = PyObject_NEW(komxx_conf_type_object, &Komxx_conf_type_type);
    r->data = new Conf_type(t);
    return (PyObject*)r;
}
